/********************************************************************************************************
		* PROGRAM      : menus
		* DATE - TIME  : mercredi 18 juillet 2007 - 14h13
		* AUTHOR       : gilles bailly ( gilles.bailly enst.fr )
			* FILENAME     : FlowerRecognizer.cpp
			* LICENSE      : GPL
			* COMMENTARY   :
********************************************************************************************************/
#include "FlowerRecognizer.h"
#include "MarkDrawing.h"


#include <sstream>
#include <iostream>
#include <fstream>
#include <math.h>

#include <QPointF>
#include "MenuGestureRecognizer.h"


float FlowerRecognizer::startI0 = 0.;
float FlowerRecognizer::startI1 = 0.55;
float FlowerRecognizer::endI0 = 0.7;
float FlowerRecognizer::endI1 = 1.0;


using namespace std;

#define MAX 1000000
#define NB_DIMENSION 4

//====================
KPlusProche::KPlusProche(): k( 5 ){ }
KPlusProche::~KPlusProche(){ }


float dist_point( QPointF& p1, QPointF& p2 ){
	return ::sqrt(  ( p1.x() - p2.x() ) * ( p1.x() - p2.x() )+ ( p1.y() - p2.y() ) * ( p1.y() - p2.y() ));
}


//================
void FlowerRecognizer::getKPPFeature( vector< QPointF >& stroke, float* feature ){
	int ind = 0;
	float angleD; float angleF; float distBoucle;
	getDimensions( stroke, angleD, angleF, distBoucle );
	
	feature[ ind++ ] = ::cos( angleD );
	feature[ ind++ ] = ::sin( angleD );
	feature[ ind++ ] = ::cos( angleF );
	feature[ ind++ ] = ::sin( angleF );
	int sumCarre, sumAbs;
	feature[ ind++ ] = getTotalAngle( stroke );
}

//=========================
float FlowerRecognizer::getTotalAngle( vector< QPointF >& stroke ){
	float res = 0;
	float xp_1, yp_1, xp, yp, alpha;
	if(stroke.size() < 4 ) { return 0; }
	for( unsigned int i = 3; i < stroke.size() - 4; i+=3 ){
		xp_1 = stroke[ i + 3 ].x() - stroke[ i ].x();
		yp_1 = stroke[ i + 3 ].y() - stroke[ i ].y();
		xp = stroke[ i ].x() - stroke[ i - 3  ].x();
		yp = stroke[ i ].y() - stroke[ i - 3  ].y();
		alpha = getAngle( xp, yp, xp_1, yp_1 );
		
		res += alpha;
	}
	
	return res;
}


//=========================
float FlowerRecognizer::getAngle( float x1, float y1, float x2, float y2 ){
	float n1 = sqrt( x1 * x1 + y1 * y1 );
	float n2 = sqrt ( x2 * x2 + y2 * y2 );
	if( n1 == 0 || n2 == 0 ) {  return 0; }
	x1 = x1 / n1; y1 /= n1;
	x2 = x2 / n2; y2 /= n2;
	float num =  ( (  x1 * x2 )  + ( y1 * y2 ) );
	float deno = 1;
	float cosAlpha;
	if( deno == 0){
		cosAlpha = 0;
	}else{
		cosAlpha = num / deno;
	}
	if( cosAlpha > 1 ){ cosAlpha = 0.99999999999999; }
	if( cosAlpha < -1 ){ cosAlpha = -0.99999999999999; }
	float alpha = ::acos( cosAlpha );

	//======= determiner le sens
	if( x1 * y2 - x2 * y1 < 0 ){ alpha = - alpha; }
	
	return alpha;
}


//==========================================
float FlowerRecognizer::getDimensions( std::vector< QPointF >& stroke, float& angleD, float& angleF, float& distBoucle ){
	
	int step0Boucle = 0;
	int step1Boucle = 0;
	int step2Boucle = 0;
	int step3Boucle = 0;
	
	int step0Angle = 0;
	int step1Angle =0;
	int step2Angle = 0;
	int step3Angle = 0;
	
	float l = 0;
	for(unsigned int i = 0; i < stroke.size() - 1; i++){
		l+= dist_point( stroke[ i ], stroke[ i+1 ] );
	}
	
	float pb0 = 0.1 * l;
	float pb1 = 0.45 * l;
	float pb2 = 0.8 * l ;
	float pb3 = 0.99 * l;
	
	float pa0 = startI0 * l;
	float pa1 = startI1 * l;
	float pa2 = endI0 * l ;
	float pa3 = endI1 * l;
	
	
	//=====================
	// recuperation de l'indice correspondant au X% de la longeur totale
	float lcPrevious = 0;
	float lc = 0 ;
	for(unsigned int i = 0; i < stroke.size() - 1; i++){
		lcPrevious = lc;
		lc += dist_point( stroke[ i ], stroke[ i+1 ] );
		if( lcPrevious < pb0  && lc > pb0  ){ step0Boucle = i; }
		if( lcPrevious < pb1  && lc > pb1  ){ step1Boucle = i; }
		if( lcPrevious < pb2  && lc > pb2  ){ step2Boucle = i; }
		if( lcPrevious < pb3  && lc > pb3  ){ step3Boucle = i; }
		
		if( lcPrevious < pa0  && lc > pa0  ){ step0Angle = i; }
		if( lcPrevious < pa1  && lc > pa1  ){ step1Angle = i; }
		if( lcPrevious < pa2  && lc > pa2  ){ step2Angle = i; }
		if( lcPrevious < pa3  && lc > pa3  ){ step3Angle = i; }
	}
	if( step3Angle == 0){ step3Angle = stroke.size() -1 ; }
	
	//====================
	// calcul angle de dÈpart et d'arrivee
	angleD = MenuGestureRecognizer::angleAtFlower( stroke[ step0Angle ].x() - stroke[ step1Angle ].x(),
			stroke[ step0Angle ].y() - stroke[ step1Angle ].y() );
	angleF = MenuGestureRecognizer::angleAtFlower( stroke[ step2Angle ].x() - stroke[ step3Angle ].x(),
			stroke[ step2Angle ].y() - stroke[ step3Angle ].y() );
	
	
	//=====================
	//recuperation des indices min pour le calcul de la distance
	float distMin = 10000;
	int iMin =0;
	int jMin=0;
	for(  int i = step0Boucle; i < step1Boucle; i++){
		for( int j = step2Boucle; j < step3Boucle; j++){
			float dist = dist_point( stroke[ i ], stroke[ j ] );
			if( dist < distMin ){
				distMin = dist;
				iMin = i;
				jMin = j;
				
			}else if ( dist == distMin ){
				if( jMin - iMin < j - i ){
					iMin = i;
					jMin = j;
					
				}
			}
		}
	}
	
	
	float lBetween = 0;
	for( int i = iMin +1; i < jMin; i++){
		lBetween += dist_point( stroke[ i-1 ], stroke[ i ] );
	}
	
	
	distBoucle = distMin / lBetween  ;
	
	
	return 0;
}




//====================
int KPlusProche::getItem(){
	int indMax = -1;
	int occurence = 0;
	
	for(int i = 0; i < k; i++){
		int tmpOccurence = 0;
		for( int j = i ; j < k; j++ ){
			if( kPlusProche[ i ].first == kPlusProche[ j ].first ){
				tmpOccurence += 1;
			}
		}
		
		if( tmpOccurence > occurence ){
			occurence = tmpOccurence;
			indMax = i;
		}
		
	}
	return kPlusProche[ indMax ].first ;
}

//==============================
FlowerRecognizer::FlowerRecognizer(){
	
}

//===============================
FlowerRecognizer::~FlowerRecognizer()
{
	for(unsigned int i= 0; i < gesturesBase.size() ; i++ ){
		if( gesturesBase[ i ] != 0 ) { delete[] gesturesBase[i]; }
	}
	
}

//===============================
//attention au nombre de composantes
void FlowerRecognizer::loadGesturesBase( const QString& nameFile){
   QFile file( nameFile );
   QString name = file.fileName();
  if( !file.exists() ){
	  qDebug( name.toLatin1() );
  }
  if (!file.open(QIODevice::ReadOnly | QIODevice::Text) )
         return;	 
  
  QTextStream in(&file);
  
  while ( !in.atEnd() ) {
         QString line = in.readLine();
			qDebug( line.toLatin1() );
			float* gestureFeature = new float[ 6 ];
		  QTextStream in2( &line );
		  qDebug( line.toLatin1() ) ;
		  in2 >> gestureFeature[ 0 ]; //numItem
		  in2 >> gestureFeature[ 1 ]; //cos D
		  in2 >> gestureFeature[ 2 ]; //sin D
		  in2 >> gestureFeature[ 3 ]; //cons F
		  in2 >> gestureFeature[ 4 ]; //sin F
		  in2 >> gestureFeature[ 5 ]; //sum Angle
		  gesturesBase.push_back( gestureFeature );
  }
  qDebug( " load gesture base ok");


/*
	ifstream file( nameFile.c_str() );
	if( file.fail() ){
		cerr<< nameFile <<" n'existe pas" <<endl;
		exit( 1 ); 	
	}
	
	
	string line;
	if( !file.eof() ){ getline( file, line ); } //header
	
	while( ! file.eof() ) {
		getline( file, line );
		std::stringstream in( line );
		float* gestureFeature = new float[ 6 ];
		
		in >> gestureFeature[ 0 ]; //numItem
		in >> gestureFeature[ 1 ]; //cos D
		in >> gestureFeature[ 2 ]; //sin D
		in >> gestureFeature[ 3 ]; //cons F
		in >> gestureFeature[ 4 ]; //sin F
		in >> gestureFeature[ 5 ]; //sum Angle
		gesturesBase.push_back( gestureFeature );
		
	}
	
	file.close();
	*/
}



//==================================
void FlowerRecognizer::getProbForGesture( float* gestureFeature, KPlusProche& result ){
	
	
	//=============
	result.kPlusProche.resize( result.k );
	for(int i = 0; i < result.k; i++ ){
		pair<int, float>  pair;
		pair.first = -1;
		pair.second = 10000;
		result.kPlusProche [ i ] = pair;
	}
	
	float dist = 0;
	for(unsigned int i = 0; i < gesturesBase.size() ; i++ ){
		dist = distanceCarre( gestureFeature, gesturesBase[ i ] );
		insert( (int) ( gesturesBase[ i ] [ 0 ] ), dist, result );
		
	}
	
}

//===============================
//cerr<<"insert "<< numItem<< " "<<dist << " "<< result.kPlusProche[ i ].second << endl;
void FlowerRecognizer::insert( int numItem, float dist, KPlusProche& result){
	
	for( int i = result.k - 1; i >= 0; i-- ){
		if( dist > result.kPlusProche[ i ].second ){
			if( i != result.k - 1 ){
				decalToRight( result.kPlusProche, i  );
				result.kPlusProche[ i + 1 ].second = dist;
				result.kPlusProche[ i + 1 ].first = numItem;
				
			}
			
			return;
		}else{
			/*
			if( dist == result.kPlusProche[ i ] .second ){
				if( result.kPlusProche[ i ].first != numItem ){
					cerr<< "kpp dist est egale "<< dist <<" " <<  result.kPlusProche[ i ] .second  << " " << numItem << result.kPlusProche[ i ].first << endl;
				}
			}
			*/
			if( i == 0 ){
				decalToRight( result.kPlusProche, -1 ); //traitement de l'insertion en tete
				result.kPlusProche[0 ].second = dist;
				result.kPlusProche[ 0 ].first = numItem;
			}
		}
	}
	
}

//===============================
//decal ind + 1 -> ind+2,  ind+2->ind+3, ind+3 ->ind+4
void FlowerRecognizer::decalToRight( vector< pair<int, float> >& v, int ind ){
	for(int i = v.size() - 2; i > ind; i-- ){
		v[ i + 1] = v[ i ];
	}
	
}


//===============================
//base gesture a une dimension en plus, c'est le numero de l'item
float FlowerRecognizer::distanceCarre( float* gestureFeature, float* featureBase ){
	float res = 0;
	
	int sens = MarkDrawing::sens( (int) featureBase[ 0 ]  );
	
	
	//=============== filtrer en fonction du sens
	if( gestureFeature[ 4 ] < 0 && featureBase[ 0 ] > 8  && sens == -1 ){
		//cerr<< "probl sens -1";
		return 10000;
	}
	
	
	
	if( gestureFeature[ 4 ] > 0 && featureBase[ 0 ] > 8 && sens == 1 ){
		//cerr<< "probl sens 1";
		return 10000;
	}
	
	
	
	
	for( int i = 0; i < 4; i++){
		res += (gestureFeature[ i ] - featureBase[ i + 1 ] ) *  (gestureFeature[ i ] - featureBase[ i + 1 ] );
	}
	
	return res;
}



